from pig_ui.ui_element import UIElement
from pig_ui.ui_textinput import UITextInput
from pig_ui.ux_element import UXWrapper, UXText, UXRect
from pig_ui.constants import *

class UISideBar(UIElement):
    """
    The SideBar is used for e.g.:
    * Settings
    * Menus
    
    Only LeftBound!
    """
    def __init__(self, app):
        size = Vector2(SCREEN_WIDTH // 4, SCREEN_HEIGHT)
        
        ux = [
            [UXRect(-1,Color('#242424' if i < 1 else '#484848'),size=size)] for i in range(4)
        ]
        
        super().__init__(app, Vector2(0,0), size, UXWrapper(ux), False,anchor = 'tl')
        # A SideBar are currently only leftbound
        # Two buttons are needed:
        #   * The first is the close buttom this is placed next to the bar (BarSize.x, 0)
        #   * The second is at 0,0
        #
        #? Only one will be drawn & update at a time.
        
        ux = [
            [UXRect(-1,Color('#242424' if i < 1 else '#484848'),size=Vector2(16,16)),
                UXText(color=Color('#969696'),text_get_callback=lambda: '<')] for i in range(4)
        ]
        
        
        self.in_btn = UIElement(
            self.app,
            Vector2(size.x,0),
            Vector2(16,16),
            UXWrapper(ux),
            parent = self,
            anchor = 'tl',
            cb_lclick = self.roll_in
        )
        
        ux = [
            [UXRect(-1,Color('#242424' if i < 1 else '#484848'),size=Vector2(16,16)),
                UXText(color=Color('#969696'),text_get_callback=lambda: '>')] for i in range(4)
        ]
        
        self.out_btn = UIElement(
            self.app,
            Vector2(0,0),
            Vector2(16,16),
            UXWrapper(ux),
            parent = self,
            anchor = 'tl',
            cb_lclick = self.roll_out
        )
        
    def roll_in(self, x):
        """
        Closes the sidebar.
        """
        self.visible = False
        self.in_btn.visible = False
        self.out_btn.visible = True
        
    def roll_out(self, x):
        """
        Opens the sidebar.
        """
        self.visible = True
        self.in_btn.visible = True
        self.out_btn.visible = False


from pig_ui.constants import *

class UX4Param:
    """
    A Parameter Object used to determine rect-boundings
    
    *options -> must be 4 Parameters
    """
    def __init__(self,*options):
        self.options = options

class UXElement:
    """
    A UXElement is used to style the UI
    """
    def draw(self):
        """
        Draws the object based on the options
        """

class UXRect(UXElement):
    """
    A Rectangle Object, will be rendered by calling draw.
    
    Uses the same options as pygame.draw.rect
    """
    def __init__(self,
                 border_radius: int | UX4Param = -1,
                 color: Color = Color('#252525'),
                 offset: Vector2 = Vector2(0,0),
                 size: Vector2 = Vector2(1,1),
                 width: int = 0):
        self.border_radius = border_radius
        if isinstance(self.border_radius, int):
            self.border_radius = UX4Param(*[self.border_radius]*4)
        self.color = color
        self.pos = offset
        self.size = size
        self.width = width
        
    def draw(self, surf: Surface, offset: Vector2):
        """
        Draws a rectangle based on the options
        """
        PG.draw.rect(surf, 
                     self.color, 
                     (offset.x + self.pos.x,offset.y + self.pos.y, self.size.x,self.size.y),
                     self.width,
                     border_radius=-1,
                     #*self.border_radius.options
                     )

class UXCircle(UXElement): 
    """
    A Circle Object, will be rendered by calling draw.
    
    Uses the same options as pygame.draw.circle
    """
    def __init__(self,
                 color: Color,
                 center: Vector2,
                 radius: float,
                 width: int,
                 drawing_points: UX4Param):
        self.color = color
        self.center = center
        self.radius = radius
        self.width = width
        self.drawing_points = drawing_points
        
    def draw(self, surf):
        """
        Draws a circle based on the options
        """
        PG.draw.circle(
            surf,
            self.color,
            self.center,
            self.radius,
            self.width,
            *self.drawing_points.options
        )

class UXLine(UXElement):
    """
    A Line Object, will be rendered by calling draw.
    
    Uses the same options as pygame.draw.line
    """
    def __init__(self,
                 color: Color,
                 start: Vector2,
                 end: Vector2,
                 width: int):
        self.color = color
        self.start = start
        self.end = end
        self.width = width
        
    def draw(self, surf):
        """
        Draws a line based on the options
        """
        PG.draw.line(
            surf,
            self.color,
            self.start,
            self.end,
            self.width
        )

class UXPolygon(UXElement):
    """
    A Polygon Object, will be rendered by calling draw.
    
    Uses the same options as pygame.draw.polygon
    """
    def __init__(self,
                 color: Color,
                 points: list[Vector2],
                 width: int):
        self.color = color
        self.points = points
        self.width = width
    def draw(self, surf):
        """
        Draws a polygon based on the options
        """
        PG.draw.polygon(
            surf,
            self.color,
            self.points,
            self.width
        )

class UXImage(UXElement):
    """
    A Image Object, this is different to the other UX.
    
    * **path**: Can be a string or Surface.(If string loads the image from the path else takes the image)
    """
    def __init__(self,
                 pos: Vector2,
                 path: str | Surface,
                 alpha: bool):
        self.pos = pos
        if isinstance(path, str):
            self.image = PG.image.load(path)
        elif isinstance(path, Surface):
            self.image = path
            
        self.alpha = alpha
        
    def update(self, surf: Surface):
        """
        Updates the image, should be the same resolution as before!
        """
        self.image = surf
        
    def draw(self, surf, offset):
        """
        Draws the object based on the options
        """
        surf.blit(self.image, self.pos+offset)

class UXText(UXElement):
    """
    A Text Object.
    Uses pygame.font to draw text
    
    * **anchor**: Should be (0=left,1=center,2=right)
    * **color**: Must be a color-like & pygame-compatible
    * **text_get_callback**: Can be str or a function(Used to change text on the fly, if necessary)
    
    Missing / Future:
    Add a snap point for x & y, so the text will not go outside the element!
    """
    def __init__(self, pos: Vector2 = Vector2(0,0), color: Color = Color('#dddddd'), anchor = 0,text_get_callback: Callable | str = lambda: ""):
        self.anchor = anchor
        self.pos = pos
        self.text_get_callback = text_get_callback
        self.color = color
        super().__init__()
        
    @property
    def anchor_offset(self) -> Vector2:
        """
        Returns the anchor offset based on `self.anchor`
        Can be:
        * 0: left
        * 0.5: center
        * 1: right
        """
        return [0,0.5,1][self.anchor]
    
    def draw(self, surf: Surface, offset: Vector2):
        """
        Renders the text(extremly inefficient) & draws it based on the options to the screen.
        """
        text = self.text_get_callback() if not isinstance(self.text_get_callback, str) else self.text_get_callback
        rendered = NORM_FONT.render(text, True, self.color)
        size = Vector2(*rendered.get_size())
        surf.blit(rendered, self.pos + offset - (size * self.anchor_offset))
        
        

UIELEMENT_DEFAULT = [
    [UXRect(-1,Color('#484848'),size=Vector2(15,15))],
    [UXRect(-1,Color('#969696'),size=Vector2(15,15))],
    [UXRect(-1,Color('#ffffff'),size=Vector2(15,15))],
    [UXRect(-1,Color('#000000'),size=Vector2(15,15))]
]


UIELEMENT_TEXT = [
    [UXText(color=Color('#484848'))],
    [UXText(color=Color('#969696'))],
    [UXText(color=Color('#ffffff'))],
    [UXText(color=Color('#000000'))]
]
            
class UXWrapper:
    """
    Wraps up the UX for UI Usage.
    """
    def __init__(self, ux: list[list[UXElement]]):
        
        self.ux = ux
        self.set_mode(0)
    
    def draw(self,surf: Surface, offset: Vector2):
        """
        Draws the UX for the selected mode
        """
        for ux in self.ux[self.selected]:
            ux.draw(surf, offset)
        
    def set_mode(self,type: int):
        """
        sets the mode, usually called by the UIElement.
        *0: normal
        *1: hover
        *2: click
        *3: disabled(unused currently)
        """
        self.selected = type
        
        



from pig_ui.ui_element import UIElement
from pig_ui.ui_dropdown import UIDropDown
from pig_ui.constants import *
from pig_ui.ux_element import UXWrapper, UXText, UXRect
from pig_ui.ui_manager import UIM
class UIMenuBar(UIElement):
    """
    A graphical element at the top-left of the screen that provides a collection of commands and options for the application.
    
    Basically a set of DropDown-Elements
    
    > [!ATTENTION]
    > You should not use multiple of this!
    """
    def __init__(self, app, pos, size, ux = None, draggable = False, **kwargs):
        kwargs['visible'] = False
        kwargs['anchor'] = 'tl'
        super().__init__(app, pos, size, ux, draggable, **kwargs)
        self.sub = []
    def set_subs(self, subs: list[dict]):
        """
        Sets the sub-elements of the UIMenu
        subs:
        * **title**: the title of the element(Is the title of a DropDown).
        * **ltext**: a list of titles for the sub-elements of the DropDown.
        * **lcom**: a list of callbacks for the sub-elements.(None will set each to lambda)
        
        Format:
        ```json
        {
            'title': "title",
            'ltext': [],
            'lcom': []
        }
        ```
        """
        
        for i, dd in enumerate(subs):
            uidd = UIDropDown(
                self.app, 
                Vector2(i*self.size.x, 0),
                self.size,
                parent=self,
                anchor= 'tl',
                title = dd['title'],
                draggable=False,
                )
            
            uidd.set_subs(dd['ltext'], dd['lcom'])
            self.sub.append(uidd)


from pig_ui.constants import *
from pig_ui.ux_element import UXWrapper, UXText, UXRect
from pig_ui.ui_element import UIElement
from pig_ui.ui_manager import UIM
class UIDropDown(UIElement):
    """
    A DropDown Menu.
    
    - * **title**: changes the title of the main element
    """
    def __init__(self, app, pos, size, **kwargs):
        kwargs['cb_lclick'] = self.toggle_dd
        kwargs['anchor'] = "tl"
        ux=UXWrapper(
            ux = [
                [UXRect(-1,Color(col),size=size),
                 UXText(text_get_callback=kwargs.get('title', ''))] for col in ('#484848', '#969696', '#ffffff', '#000000')
            ]
        )
        super().__init__(app, pos, size, ux, False, **kwargs)
        self.sub = []

    @property
    def text(self) -> str:
        """
        !Unused
        """
        return "test"
    
    def toggle_dd(self,*_):
        """
        Toggles the visibility of all SubElements
        """
        for uie in self.sub:
            uie.visible = not uie.visible
            
    def sub_callbacks(self,obj):
        """
        Runs the corresponding callback for the dropdown subelement.
        Determined by the object
        """
        i = self.sub.index(obj)
        self.sub_cbs[i](obj)
        for uie in self.sub:
            uie.visible = not uie.visible
            
    def set_subs(self, ltext: list[str], lcom: list[Callable] | None = None):
        """
        Set SubElements & Resets existing Subs if exist:
        Format:
        ltext:
        ```json
        [
            "abc",
            "def",
            ...
        ]
        lcom:
        None or
        [
            function,
            lambda x: print(x),
            ...
        ]
        ```
        """
        for uie in self.sub:
            uie: UIElement
            uie.destroy()
        
        self.texts = []
        if lcom is None:
            lcom = [lambda *x: None for i in ltext]
        self.sub_cbs = lcom
        for i, (t, c) in enumerate(zip(ltext, lcom)):
            self.texts.append(t)
            ux = [
                [UXRect(-1,Color(col),size=self.size),
                 UXText(text_get_callback=t)] for col in ('#484848', '#969696', '#ffffff', '#000000')
            ]
            
            uie = UIElement(
                self.app, 
                Vector2(0,(i+1) * self.size.y),
                self.size,
                draggable=False,
                ux=UXWrapper(ux), 
                parent=self,
                anchor="tl",
                cb_lclick=self.sub_callbacks,
                cb_dclick=lambda x: None,
                visible=False
                )
            self.sub.append(uie)


from pig_ui.constants import *

def draw_beziere(screen, control_points, numPoints = None,color=Color('#ffffff'), width=1):
    """
    Draws a smooth cubic Bezier curve onto the specified screen surface.
    """
    control_points = [Vector2(f,g) for f,g in control_points]
    b_points = compute_bezier_points([(x.x, x.y) for x in control_points])
    PG.draw.lines(screen, color, False, b_points, width=width)

def compute_bezier_points(vertices, numPoints=None):
    """
    Computes a set of discrete points along a cubic Bezier curve.
    """
    if numPoints is None:
        numPoints = 30
    if numPoints <= 2 or len(vertices) != 4:
        return None

    result = []

    b0x = vertices[0][0]
    b0y = vertices[0][1]
    b1x = vertices[1][0]
    b1y = vertices[1][1]
    b2x = vertices[2][0]
    b2y = vertices[2][1]
    b3x = vertices[3][0]
    b3y = vertices[3][1]



    # Compute polynomial coefficients from Bezier points
    ax = -b0x + 3 * b1x + -3 * b2x + b3x
    ay = -b0y + 3 * b1y + -3 * b2y + b3y

    bx = 3 * b0x + -6 * b1x + 3 * b2x
    by = 3 * b0y + -6 * b1y + 3 * b2y

    cx = -3 * b0x + 3 * b1x
    cy = -3 * b0y + 3 * b1y

    dx = b0x
    dy = b0y

    # Set up the number of steps and step size
    numSteps = numPoints - 1 # arbitrary choice
    h = 1.0 / numSteps # compute our step size

    # Compute forward differences from Bezier points and "h"
    pointX = dx
    pointY = dy

    firstFDX = ax * (h * h * h) + bx * (h * h) + cx * h
    firstFDY = ay * (h * h * h) + by * (h * h) + cy * h


    secondFDX = 6 * ax * (h * h * h) + 2 * bx * (h * h)
    secondFDY = 6 * ay * (h * h * h) + 2 * by * (h * h)

    thirdFDX = 6 * ax * (h * h * h)
    thirdFDY = 6 * ay * (h * h * h)

    # Compute points at each step
    result.append((int(pointX), int(pointY)))

    for i in range(numSteps):
        pointX += firstFDX
        pointY += firstFDY

        firstFDX += secondFDX
        firstFDY += secondFDY

        secondFDX += thirdFDX
        secondFDY += thirdFDY

        result.append((int(pointX), int(pointY)))

    return result
